@@ -1,11 +1,12 @@
 ;;; buttercup.el --- Behavior-Driven Emacs Lisp Testing -*-lexical-binding:t-*-
 
 ;; Copyright (C) 2015-2017  Jorgen Schaefer <contact@jorgenschaefer.de>
+;; Copyright (C) 2018-2024  Ola Nilsson <ola.nilsson@gmail.com>
 
-;; Version: 1.31
+;; Version: 1.37
 ;; Author: Jorgen Schaefer <contact@jorgenschaefer.de>
 ;; Maintainer: Ola Nilsson <ola.nilsson@gmail.com>
-;; Package-Requires: ((emacs "24.3"))
+;; Package-Requires: ((emacs "24.4"))
 ;; URL: https://github.com/jorgenschaefer/emacs-buttercup
 
 ;; This program is free software; you can redistribute it and/or
@@ -31,6 +32,7 @@
 ;; the first parameter describing the suite and the rest being the body
 ;; of code that implements the suite.
 
+;; ;;; lexical binding is required -*- lexical-binding: t; -*-
 ;; (describe "A suite"
 ;;   (it "contains a spec with an expectation"
 ;;     (expect t :to-be t)))
@@ -54,6 +56,9 @@
 ;; inconsistencies.
 (define-error 'buttercup-internals-error "Internal buttercup error"
               'buttercup-error-base)
+;; Raised when expanding `describe` macros whithout lexical-binding: t
+(define-error 'buttercup-dynamic-binding-error "Lexical binding is not enabled"
+              'buttercup-error-base)
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; wrapper function manipulation
@@ -63,19 +68,29 @@
               "Bad test expression"
               'buttercup-internals-error)
 
-(defun buttercup--enclosed-expr (fun)
-  "Given a zero-arg function FUN, return its unevaluated expression.
+(eval-and-compile
+  (if (fboundp 'oclosure-define)      ;Emacs≥29
+    (oclosure-define (buttercup--thunk (:predicate buttercup--thunk-p))
+      "An elisp expression as a function and original code."
+      expr)
+    (defalias 'buttercup--thunk-p #'ignore
+      "Always return nil when Oclosures are not available.")))
 
+(defun buttercup--enclosed-expr (fun)
+  "Given a FUN `buttercup-thunk', return its unevaluated expression.
+
+For Emacs < 29:
 The function MUST be byte-compiled or have one of the following
 forms:
 
-\(closure (ENVLIST) () (quote EXPR) (buttercup--mark-stackframe) EXPANDED)
-\(lambda () (quote EXPR) (buttercup--mark-stackframe) EXPR)
+\(closure (ENVLIST) () (quote EXPR) EXPANDED)
+\(lambda () (quote EXPR) EXPR)
 
 and the return value will be EXPR, unevaluated. The quoted EXPR
 is useful if EXPR is a macro call, in which case the `quote'
 ensures access to the un-expanded form."
-  (cl-assert (functionp fun) t "Expected FUN to be a function")
+  (if (buttercup--thunk-p fun)
+      (buttercup--thunk--expr fun)
   (pcase fun
     ;; This should be the normal case, a closure with unknown enclosed
     ;; variables, empty arglist and a body containing
@@ -83,7 +98,7 @@ ensures access to the un-expanded form."
     ;; * the stackframe marker
     ;; * the macroexpanded original expression
     (`(closure ,(pred listp) nil
-        (quote ,expr) (buttercup--mark-stackframe) ,_expanded)
+        (quote ,expr) ,_expanded)
      expr)
     ;; This a when FUN has not been evaluated.
     ;; Why does that happen?
@@ -92,17 +107,17 @@ ensures access to the un-expanded form."
     ;; * the stackframe marker
     ;; * the expanded expression
     (`(lambda nil
-        (quote ,expr) (buttercup--mark-stackframe) ,_expanded)
+        (quote ,expr) ,_expanded)
      expr)
-    ;;; This is when FUN has been byte compiled, as when the entire
-    ;;; test file has been byte compiled. Check that it has an empty
-    ;;; arglist, that is all that is possible at this point. The
-    ;;; return value is byte compiled code, not the original
-    ;;; expressions. Also what is possible at this point.
+    ;; This is when FUN has been byte compiled, as when the entire
+    ;; test file has been byte compiled. Check that it has an empty
+    ;; arglist, that is all that is possible at this point. The
+    ;; return value is byte compiled code, not the original
+    ;; expressions. Also what is possible at this point.
     ((and (pred byte-code-function-p) (guard (member (aref fun 0) '(nil 0))))
      (aref fun 1))
     ;; Error
-    (_ (signal 'buttercup-enclosed-expression-error (format "Not a zero-arg one-expression closure: %S" fun)))))
+    (_ (signal 'buttercup-enclosed-expression-error (format "Not a zero-arg one-expression closure: %S" fun))))))
 
 (defun buttercup--expr-and-value (fun)
   "Given a function, return its quoted expression and value.
@@ -170,26 +185,31 @@ Does not have the IGNORE-MISSING and SPLIT parameters."
 (define-error 'buttercup-pending "Buttercup test is pending" 'buttercup-error-base)
 
 (defun buttercup--wrap-expr (expr)
-  "Wrap EXPR to be used by `buttercup-expect'."
-  `(lambda ()
-     (quote ,expr)
-     (buttercup--mark-stackframe)
-     ,expr))
+  "Wrap EXPR in a `buttercup--thunk' to be used by `buttercup-expect'.
+This function is only usable from within the buttercup `expect' macro."
+  (if (fboundp 'oclosure-lambda)        ;Emacs≥29
+      `(oclosure-lambda (buttercup--thunk (expr ',expr)) ()
+         ,expr)
+    `(lambda ()
+       (quote ,expr)
+       ,expr)))
 
 (defmacro expect (arg &optional matcher &rest args)
   "Expect a condition to be true.
 
-This macro knows three forms:
+This macro knows two forms:
 
-\(expect ARG :MATCHER ARGS...)
-  Fail the current test if the matcher does not match these arguments.
+\(expect ARG :MATCHER [ARGS...])
+  Fail the current test if the MATCHER does not match these arguments.
+  The correct number of arguments depend on the MATCHER.
   See `buttercup-define-matcher' for more information on matchers.
 
-\(expect (function ARG...))
-  Fail the current test if the function call does not return a true value.
-
 \(expect ARG)
-  Fail the current test if ARG is not true."
+  Fail the current test if ARG is not true.
+  This is the same as (expect ARG :to-be-truthy).
+
+All arguments can be a value or a single Lisp form. The evaluation of
+the arguments is delayed until the containing spec is executed."
   `(buttercup-expect ,(buttercup--wrap-expr arg)
                      ,(or matcher :to-be-truthy)
                      ,@(mapcar #'buttercup--wrap-expr args)))
@@ -200,24 +220,14 @@ This macro knows three forms:
 See the macro documentation for details and the definition of
 ARG, MATCHER and ARGS."
   (cl-assert (cl-every #'buttercup--wrapper-fun-p (cons arg args)) t)
-  (if (not matcher)
-      (progn
-        (cl-assert (not args) t)
-        (when (not (funcall arg))
-          (buttercup-fail "Expected %S to be non-nil"
-                          (buttercup--enclosed-expr arg))))
-    (let ((result (buttercup--apply-matcher matcher (cons arg args))))
-      (if (consp result)
-          (when (not (car result))
-            (buttercup-fail "%s" (cdr result)))
-        (when (not result)
-          (buttercup-fail "Expected %S %S %s"
-                          (buttercup--enclosed-expr arg)
-                          matcher
-                          (mapconcat (lambda (obj)
-                                       (format "%S" (funcall obj)))
-                                     args
-                                     " ")))))))
+  (pcase (buttercup--apply-matcher (or matcher :to-be-truthy) (cons arg args))
+    (`(,result . ,message) (unless result (buttercup-fail message)))
+    (result (unless result (buttercup-fail "Expected %S %S %s"
+                                           (buttercup--enclosed-expr arg)
+                                           matcher
+                                           (mapconcat (lambda (obj)
+                                                        (format "%S" (funcall obj)))
+                                                      args " "))))))
 
 (defun buttercup-fail (format &rest args)
   "Fail the current test with the given description.
@@ -246,16 +256,16 @@ MESSAGE is omitted or nil show the condition form instead."
 
 (defmacro buttercup-define-matcher (matcher args &rest body)
   "Define a matcher named MATCHER to be used in `expect'.
-
+MATCHER is a keyword, for instance `:to-be'.
 ARGS is a list of the elements to match together.
-
 The BODY will receive ARGS as functions that can be called (using
 `funcall') to get their values. BODY should return either a
 simple boolean, or a cons cell of the form (RESULT . MESSAGE). If
 RESULT is nil, MESSAGE should describe why the matcher failed. If
 RESULT is non-nil, MESSAGE should describe why a negated matcher
 failed.
-BODY may start with a docstring."
+
+\(fn MATCHER ARGS [DOCSTRING] BODY...)"
   (declare (indent defun))
   `(put ,matcher 'buttercup-matcher
         (lambda ,args
@@ -597,6 +607,8 @@ See also `buttercup-define-matcher'."
   :expect-mismatch-phrase "Expected `%A' < %b, but `%A' was %a.")
 
 (buttercup-define-matcher :to-be-close-to (a b precision)
+  "Verify that |A-B| < 10^-PRECISION.
+\(expect A :to-be-close-to B PRECISION)"
   (cl-destructuring-bind
       (precision (a-expr . a) (_b-expr . b))
       (cons (funcall precision)
@@ -656,58 +668,79 @@ UNEVALUATED-EXPR is the Lisp sexp used before the :to-throw
 matcher keyword in the `expect' statement.
 EXPR-VALUE is the return value from the evaluation of
 UNEVALUATED-EXPR if it did not raise any signal."
-  (let ((thrown-signal-symbol (car thrown-signal))
-        (thrown-signal-args (cdr thrown-signal))
-        (expected-signal-symbol (car expected-signal))
-        (expected-signal-args (cdr expected-signal)))
+  (let* ((thrown-signal-symbol (car thrown-signal))
+         (thrown-signal-args (cdr thrown-signal))
+         (expected-signal-symbol (car expected-signal))
+         (expected-signal-args (cdr expected-signal))
+         (matching-signal-symbol
+          (or (null expected-signal-symbol)
+              (memq expected-signal-symbol (get thrown-signal-symbol 'error-conditions))))
+         (explained-signal-args ; nil for matched, explained or t for mismatched
+          (when expected-signal-args
+              ;; The ert-explainer for equal does an equal internally,
+              ;; so avoid calling equal twice by calling the explainer
+              ;; directly.
+              (funcall (or (get 'equal 'ert-explainer) (lambda (a b) (not (equal a b))))
+                       thrown-signal-args expected-signal-args))))
     (let*
-        ((matched
-          (and thrown-signal
-               (or (null expected-signal-symbol)
-                   (memq expected-signal-symbol (get thrown-signal-symbol 'error-conditions)))
-               (or (null expected-signal-args)
-                   (equal thrown-signal-args expected-signal-args))))
+        ((matched (and thrown-signal matching-signal-symbol (not explained-signal-args)))
          ;; Some of these replacement are always used, there is no
          ;; reason not to format them immediately. But e and t are not
          ;; always used and should be delayed. Use
          ;; buttercup--simple-format for formatting as format-spec
          ;; does not support functions until Emacs 29
          (spec (format-spec-make
-                ?E (format "%S" unevaluated-expr)
-                ?e (lambda () (format "%S" expr-value))
+                ?E (format "`%S'" unevaluated-expr)
+                ?e (lambda () (format "`%S'" expr-value))
                 ?t (lambda () (format "%S" thrown-signal))
-                ?S (lambda () (format "%S" thrown-signal-symbol))
-                ?A (lambda () (format "%S" thrown-signal-args))
+                ?S (lambda () (format "`%S'" thrown-signal-symbol))
+                ?A (lambda ()
+                     (if expected-signal-args
+                         (format " with args `%S'" thrown-signal-args)
+                       ""))
                 ?s (if expected-signal-symbol
                        (format "a child signal of `%S'" expected-signal-symbol)
                      "a signal")
                 ?a (if expected-signal-args
                        (format " with args `%S'" expected-signal-args)
-                     "")))
-         (result-fmt (if thrown-signal
-                         "it threw `%S' with args %A"
-                       "it evaluated successfully, returning value `%e'")))
-      (buttercup--test-expectation matched
-        :expect-match-phrase
-        (buttercup--simple-format
-         spec
-         "Expected `%E' to throw %s%a, but instead " result-fmt)
-        :expect-mismatch-phrase
-        (buttercup--simple-format
-         spec
-         "Expected `%E' not to throw %s%a, but " result-fmt)))))
+                     "")
+                ?q (lambda () (format "%S" explained-signal-args))
+                )))
+      (cond
+       (matched ;; should be the most likely result
+        `(t . ,(buttercup--simple-format
+                spec "Expected %E not to throw %s%a, but it threw %S%A")))
+       ((null thrown-signal) ; no signal raised
+        `(nil . ,(buttercup--simple-format
+                  spec "Expected %E to throw %s%a, but instead it returned %e")))
+       ((and explained-signal-args (not matching-signal-symbol)) ; neither symbol nor args matched
+        `(nil . ,(buttercup--simple-format
+                  spec
+                  "Expected %E to throw %s%a, but instead it threw %S%A")))
+       (explained-signal-args ; symbol matched
+        `(nil . ,(buttercup--simple-format
+                  spec
+                  "Expected %E to signal %s%a, but instead signalled%A which does not match because %q.")))
+       ((not matching-signal-symbol) ; args matched
+        `(nil . ,(buttercup--simple-format
+                  spec
+                  "Expected %E to throw %s%a, but instead it threw %S%A")))
+       (t (error "`buttercup--handle-to-throw' could not handle args %S %S"
+                 thrown-signal expected-signal))))))
 
 (buttercup-define-matcher :to-have-been-called (spy)
+    "Check that SPY have been called at least once."
   (cl-assert (symbolp (funcall spy)))
   (if (spy-calls-all (funcall spy))
       t
     nil))
 
 (buttercup-define-matcher :to-have-been-called-with (spy &rest args)
+  "Check that SPY has been called at least once with arguments ARGS."
   (setq spy (funcall spy))
   (cl-assert (symbolp spy))
   (setq args (mapcar #'funcall args))
-  (let* ((calls (mapcar 'spy-context-args (spy-calls-all spy))))
+  (let* ((calls (mapcar #'spy-context-args (spy-calls-all spy))))
     (cond
      ((not calls)
       (cons nil
@@ -725,13 +758,16 @@ UNEVALUATED-EXPR if it did not raise any signal."
       t))))
 
 (buttercup-define-matcher :to-have-been-called-times (spy number)
+  "Check that SPY has been called exactly NUMBER times."
   (setq spy (funcall spy)
         number (funcall number))
   (cl-assert (symbolp spy))
   (let* ((call-count (spy-calls-count spy)))
     (cond
      ((= number call-count)
-      t)
+      (cons t
+            (format "Expected `%s' to not have been called exactly %s %s, but it was."
+                    spy number (if (= number 1) "time" "times"))))
      (t
       (cons nil
             (format "Expected `%s' to have been called %s %s, but it was called %s %s"
@@ -784,8 +820,10 @@ Return CHILD."
     (cons (buttercup-suite-or-spec-parent suite-or-spec)
           (buttercup-suite-or-spec-parents (buttercup-suite-or-spec-parent suite-or-spec)))))
 
-(define-obsolete-function-alias 'buttercup-suite-parents 'buttercup-suite-or-spec-parents "emacs-buttercup 1.10")
-(define-obsolete-function-alias 'buttercup-spec-parents 'buttercup-suite-or-spec-parents "emacs-buttercup 1.10")
+(define-obsolete-function-alias 'buttercup-suite-parents
+  #'buttercup-suite-or-spec-parents "emacs-buttercup 1.10")
+(define-obsolete-function-alias 'buttercup-spec-parents
+  #'buttercup-suite-or-spec-parents "emacs-buttercup 1.10")
 
 (defun buttercup-suites-total-specs-defined (suite-list)
   "Return the number of specs defined in all suites in SUITE-LIST."
@@ -910,6 +948,9 @@ form.")
 DESCRIPTION is a string. BODY is a sequence of instructions,
 mainly calls to `describe', `it' and `before-each'."
   (declare (indent 1) (debug (&define sexp def-body)))
+  (unless lexical-binding
+    (signal 'buttercup-dynamic-binding-error
+            "buttercup requires `lexical-binding' to be t"))
   (let ((new-body
          (cond
           ((eq (elt body 0) :var)
@@ -967,7 +1008,6 @@ most probably including one or more calls to `expect'."
       `(buttercup-it ,description
          (lambda ()
            (buttercup-with-converted-ert-signals
-             (buttercup--mark-stackframe)
              ,@body)))
     `(buttercup-xit ,description)))
 
@@ -1345,11 +1385,11 @@ in a `buttercup-with-cleanup' environment.")
 
 (defun spy-calls-count-returned (spy)
   "Return the number of times SPY has been called successfully so far."
-  (cl-count-if 'spy-context-return-p (spy-calls-all spy)))
+  (cl-count-if #'spy-context-return-p (spy-calls-all spy)))
 
 (defun spy-calls-count-errors (spy)
   "Return the number of times SPY has been called and thrown errors so far."
-  (cl-count-if 'spy-context-thrown-p (spy-calls-all spy)))
+  (cl-count-if #'spy-context-thrown-p (spy-calls-all spy)))
 
 (defun spy-calls-args-for (spy index)
   "Return the context of the INDEXth call to SPY."
@@ -1361,7 +1401,7 @@ in a `buttercup-with-cleanup' environment.")
 
 (defun spy-calls-all-args (spy)
   "Return the arguments for every recorded call to SPY."
-  (mapcar 'spy-context-args (spy-calls-all spy)))
+  (mapcar #'spy-context-args (spy-calls-all spy)))
 
 (defun spy-calls-most-recent (spy)
   "Return the context of the most recent call to SPY."
@@ -1413,8 +1453,11 @@ each frame, and prefixes each stack frame with lambda or M to
 indicate whether it represents a normal evaluated function call
 or a macro/special form.")
 
-(defvar buttercup-color t
-  "Whether to use colors in output.")
+(defvar buttercup-color (let ((no-color (getenv "NO_COLOR")))
+                          (or (not no-color) (string= no-color "")))
+  "Whether to use colors in output.
+Will be nil if the `NO_COLOR' environment variable is set to any value
+\(see URL https://no-color.org/).")
 
 (defconst buttercup-warning-buffer-name " *Buttercup-Warnings*"
   "Buffer name used to collect warnings issued while running a spec.
@@ -1483,7 +1526,7 @@ current directory."
         (setq args (cdr args)))))
     (setq command-line-args-left nil)
     (setq failed-files-suite (make-buttercup-suite
-                              :description "File failed to load completely: "))
+                              :description "File failed to load correctly:"))
     (dolist (dir (or dirs '(".")))
       (dolist (file (directory-files-recursively
                      dir "\\`test-.*\\.el\\'\\|-tests?\\.el\\'"))
@@ -1506,9 +1549,12 @@ spec and add it to FAILURE-SUITE."
   (cl-destructuring-bind (status description stack)
       (buttercup--funcall #'load file nil t)
     (when (eq status 'failed)
+      (if (memq (cl-caadr description)
+                '(buttercup-dynamic-binding-error end-of-file))
+          (setq stack nil)
       ;; Skip all of stack until load is called
       (while (not (eq (nth 1 (car stack)) 'load))
-        (pop stack))
+        (pop stack)))
       (buttercup-suite-add-child
        failure-suite
        (make-buttercup-spec
@@ -1696,7 +1742,7 @@ Do not change the global value.")
   "Update SUITE-OR-SPEC with the result of calling FUNCTION with ARGS.
 Sets the `status', `failure-description', and `failure-stack' for
 failed and pending specs."
-  (let* ((result (apply 'buttercup--funcall function args))
+  (let* ((result (apply #'buttercup--funcall function args))
          (status (elt result 0))
          (description (elt result 1))
          (stack (elt result 2)))
@@ -1766,6 +1812,15 @@ Two special statuses can be listed in
            (string= (buttercup-spec-failure-description spec) "PENDING"))
       ))
 
+(defun buttercup--reporter-batch-preprint-spec-p (spec)
+  "Return non-nil if the SPEC description should be printed at `spec-started'."
+  (not (or buttercup-reporter-batch-quiet-statuses
+           ;; Do not 'pre-print' in github actions unless color is
+           ;; disabled. See #181.
+           (and (getenv "GITHUB_ACTION") buttercup-color)
+           (and buttercup-color
+                (string-match-p "[\n\v\f]" (buttercup-spec-description spec))))))
+
 (defun buttercup-reporter-batch (event arg)
   "A reporter that handles batch sessions.
 
@@ -1790,18 +1845,23 @@ EVENT and ARG are described in `buttercup-reporter'."
            (push arg buttercup-reporter-batch--suite-stack)
          (buttercup--print "%s\n" (buttercup--indented-description arg))))
       (`spec-started
-       (or buttercup-reporter-batch-quiet-statuses
-           (and buttercup-color
-                (string-match-p "[\n\v\f]" (buttercup-spec-description arg)))
-           (buttercup--print "%s" (buttercup--indented-description arg))))
+       (when (buttercup--reporter-batch-preprint-spec-p arg)
+         (buttercup--print "%s" (buttercup--indented-description arg))))
       (`spec-done
+       ;; When printing has been held back but we will print the
+       ;; result of this spec, print the all of the containing suite
+       ;; descriptions that have been held back
        (when (and buttercup-reporter-batch-quiet-statuses
                   (not (buttercup-reporter-batch--quiet-spec-p arg)))
          (dolist (suite (nreverse buttercup-reporter-batch--suite-stack))
            (buttercup--print "%s\n" (buttercup--indented-description suite)))
          (setq buttercup-reporter-batch--suite-stack nil)
-         (buttercup--print "%s" (buttercup--indented-description arg)))
+         ;; Also print the spec description unless it should not be preprinted
+         (unless (buttercup--reporter-batch-preprint-spec-p arg)
+           (buttercup--print "%s" (buttercup--indented-description arg))))
 
+       ;; print the result of the spec. This should erase any
+       ;; non-colored spec text.
        (unless (buttercup-reporter-batch--quiet-spec-p arg)
          (buttercup-reporter-batch--print-spec-done-line arg buttercup-color))
 
@@ -1829,8 +1889,8 @@ EVENT and ARG are described in `buttercup-reporter'."
 (defun buttercup-reporter-batch--print-spec-done-line (spec color)
   "Print the remainder of the SPEC report line for `spec-done'.
 
-If COLOR is non-nil, erace the text so far on the current line
-using '\r' and replace it with the same text colored according to
+If COLOR is non-nil, erase the text so far on the current line
+using '\\r' and replace it with the same text colored according to
 the SPEC status. Do not erase and replace if the text would have
 been reprinted with the default color.
 
@@ -1861,10 +1921,14 @@ Finally print the elapsed time for SPEC."
                              (`failed 'red)
                              (`skipped nil))))
     (when color
-      ;; Carriage returns (\r) should not be colorized. It would mess
-      ;; up color handling in Emacs compilation buffers using
-      ;; `ansi-color-apply-on-region' in `compilation-filter-hook'.
-      (buttercup--print "\r%s"
+      ;; Clear the line if
+      (when (or buttercup-reporter-batch-quiet-statuses
+                (buttercup--reporter-batch-preprint-spec-p spec))
+        ;; Carriage returns (\r) should not be colorized. It would mess
+        ;; up color handling in Emacs compilation buffers using
+        ;; `ansi-color-apply-on-region' in `compilation-filter-hook'.
+        (buttercup--print "\r"))
+      (buttercup--print "%s"
                         (buttercup-colorize
                          (buttercup--indented-description spec) color)))
     (unless (eq 'passed status)
@@ -1930,8 +1994,7 @@ Colorize parts of the output if COLOR is non-nil."
 FMT and ARGS are passed to `format'."
   (send-string-to-terminal (apply #'format fmt args)))
 
-
-(defadvice display-warning (around buttercup-defer-warnings activate)
+(defun buttercup--display-warning (fn type message &optional level buffer-name &rest args)
   "Log all warnings to a special buffer while running buttercup specs.
 
 Emacs' normal display logic for warnings doesn't mix well with
@@ -1951,8 +2014,10 @@ finishes."
       (cl-letf
           ((warning-minimum-level :emergency)
            ((symbol-function 'message) 'ignore))
-        ad-do-it)
-    ad-do-it))
+        (apply fn type message level buffer-name args))
+    (apply fn type message level buffer-name args)))
+
+(advice-add 'display-warning :around #'buttercup--display-warning)
 
 (defconst buttercup-colors
   '((black   . 30)
@@ -1980,7 +2045,7 @@ the capturing behavior."
 (defun buttercup-colorize (string color)
   "Format STRING with COLOR.
 Return STRING unmodified if COLOR is nil."
-  (if color
+  (if (and color buttercup-color)
       (let ((color-code (cdr (assq color buttercup-colors))))
         (format "\e[%sm%s\e[0m" color-code string))
     string))
@@ -2002,7 +2067,7 @@ EVENT and ARG are described in `buttercup-reporter'."
                                 (with-current-buffer buf
                                   (let ((inhibit-read-only t))
                                     (goto-char (point-max))
-                                    (insert (apply 'format fmt args))))))
+                                    (insert (apply #'format fmt args))))))
       (unwind-protect
           (let ((buttercup-color))
             (buttercup-reporter-batch event arg))
@@ -2035,52 +2100,96 @@ failed -- The second value is the description of the expectation
 
 (defun buttercup--debugger (&rest args)
   "Debugger function that return error context with an exception.
-
 ARGS according to `debugger'."
   ;; If we do not do this, Emacs will not run this handler on
   ;; subsequent calls. Thanks to ert for this.
   (setq num-nonmacro-input-events (1+ num-nonmacro-input-events))
   (throw 'buttercup-debugger-continue
          (list 'failed args
-               (cl-destructuring-bind (_ (signal-type . data)) args
-                 (unless (eq signal-type 'buttercup-pending)
-                   (buttercup--backtrace))))))
-
-(defalias 'buttercup--mark-stackframe 'ignore
-  "Marker to find where the backtrace start.")
+               ;; args is (error (signal . data) ....) where the tail
+               ;; may be empty
+               (cl-destructuring-bind (signal-type . data) (cl-second args)
+                 (cl-case signal-type
+                   ((buttercup-pending buttercup-failed))
+                   (otherwise (buttercup--backtrace)))))))
 
 (defun buttercup--backtrace ()
   "Create a backtrace, a list of frames returned from `backtrace-frame'."
-  ;; Read the backtrace frames from 0 (the closest) upward.
-  (cl-do* ((n 0 (1+ n))
-           (frame (backtrace-frame n) (backtrace-frame n))
-           (frame-list nil)
-           (in-program-stack nil))
+  ;; Read the backtrace frames from `buttercup--debugger' + 1  upward.
+  (cl-do* ((n 1 (1+ n))
+           (frame (backtrace-frame n #'buttercup--debugger)
+                  (backtrace-frame n #'buttercup--debugger))
+           (frame-list nil))
       ((not frame) frame-list)
-      ;; discard frames until (and including) `buttercup--debugger', they
-      ;; only contain buttercup code
-      (when in-program-stack
-        (push frame frame-list))
-      (when (eq (elt frame 1)
-                'buttercup--debugger)
-        (setq in-program-stack t))
-      ;; keep frames until one of the known functions are found, after
-      ;; this is just the buttercup framework and not interesting for
-      ;; users incorrect for testing buttercup. Some frames before the
-      ;; function also have to be discarded
-      (cl-labels ((tree-find (key tree)
-                             (cl-block tree-find
-                               (while (consp tree)
-                                 (let ((elem (pop tree)))
-                                   (when (or (and (consp elem)
-                                                  (tree-find key elem))
-                                             (eql key elem))
-                                     (cl-return-from tree-find t))))
-                               (cl-return-from tree-find
-                                 (and tree (eql tree key))))))
-        (when (and in-program-stack (tree-find 'buttercup--mark-stackframe frame))
-          (pop frame-list)
-          (cl-return frame-list)))))
+    ;; Keep frames until one if the end conditions is met. After
+    ;; this is just the buttercup framework and not interesting for
+    ;; users - except for testing buttercup.
+    (when (or
+           ;; When the error occurs in the calling of one of the
+           ;; wrapped expressions of an expect.
+           (buttercup--wrapper-fun-p (cadr frame))
+           ;; When an error happens in spec code but outside an expect
+           ;; statement
+           ;; buttercup--update-with-funcall
+           ;;  apply buttercup--funcall
+           ;;   buttercup--funcall   -   sets debugger
+           ;;    apply FUNCTION
+           ;;     FUNCTION  -- spec body function
+           ;;      condition-case   -- from buttercup-with-converted-ert-signals
+           ;;       (let ((buttercup--stackframe-marker 1))  -- the same
+           ;;        ACTUAL CODE
+           (and (null (car frame))
+                (eq 'let (cadr frame))
+                (equal '((buttercup--stackframe-marker 1)) (car (cddr frame)))
+                )
+           ;; TODO: What about :to-throw?
+           ;; buttercup--update-with-funcall (spec ...
+           ;;  apply buttercup--funcall
+           ;;   buttercup--funcall   -- sets the debugger
+           ;;    apply FUNCTION
+           ;;     FUNCTION -- spec body function
+           ;;      condition-case -- from buttercup-with-converted-ert-signals
+           ;;       (let ((buttercup--stackframe-marker 1))
+           ;;        (buttercup-expect
+           ;;         (buttercup--apply-matcher
+           ;;          (apply to-throw-matcher
+           ;;           (to-throw-matcher
+           ;;             We need a new debugger here, the
+           ;;             condition-case can not be used to collect
+           ;;             backtrace.
+           ;; When the error happens in the matcher function
+           ;;  (buttercup-expect
+           ;;   (buttercup--apply-matcher
+           ;;    (apply some-kind-of-function
+           ;;     (matcher
+           ;;      ACTUAL CODE
+           (and (eq 'buttercup--apply-matcher (cadr frame))
+                ;; The two preceeding frames are not of user interest
+                (pop frame-list) (pop frame-list)
+                ;; Add a fake frame for the matcher function
+                (push (cons t
+                            (cons (car (cddr frame))
+                                  (mapcar (lambda (x)
+                                            (if (buttercup--wrapper-fun-p x)
+                                                (buttercup--enclosed-expr x)
+                                              x))
+                                          (cadr (cddr frame)))))
+                      frame-list))
+           ;; TODO: What about signals in before and after blocks?
+           ;; BEFORE-EACH:
+           ;; buttercup--run-suite
+           ;;  (let* ...
+           ;;   (dolist (f (buttercup-suite-before-all ...
+           ;;    (buttercup--update-with-funcall suite f
+           ;;     (apply buttercup--funcall
+           ;;      (buttercup-funcall f
+           ;;       (f)
+           ;; Currently, buttercup silently ignores error in
+           ;; (before|after)-(all|each). As long as that is the case,
+           ;; there is nothing we can do about stacktraces.
+           )
+      (cl-return frame-list))
+    (push frame frame-list)))
 
 (defun buttercup--format-stack-frame (frame &optional style)
   "Format stack FRAME according to STYLE.
@@ -2129,7 +2238,9 @@ Specifically, `ert-test-failed' is converted to
 `buttercup-pending'."
   (declare (indent 0))
   `(condition-case err
-       (progn ,@body)
+       (let ((buttercup--stackframe-marker 1))
+         (ignore buttercup--stackframe-marker)
+         ,@body)
      (ert-test-failed
       (buttercup-fail "%S" err))
      (ert-test-skipped
@@ -2187,10 +2298,9 @@ ARG is ‘toggle’; disable otherwise."
         (t (remove-hook 'after-load-functions #'buttercup-check-for-stale-elc))))
 
 (provide 'buttercup)
-;;; buttercup.el ends here
-
 ;; Local Variables:
 ;; indent-tabs-mode: nil
 ;; sentence-end-double-space: nil
 ;; tab-width: 8
 ;; End:
+;;; buttercup.el ends here
